// -*- C++ -*-
/*!
 * @file  DataConversionRTC.cpp
 * @brief DataConversionRTC
 * @date $Date$
 *
 * $Id$
 */

#include "DataConversionRTC.h"

// Module specification
// <rtc-template block="module_spec">
static const char* dataconversionrtc_spec[] =
  {
    "implementation_id", "DataConversionRTC",
    "type_name",         "DataConversionRTC",
    "description",       "DataConversionRTC",
    "version",           "1.0.0",
    "vendor",            "TakahashiNaoki",
    "category",          "Three-WheeledMobileRobot",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.PGainL", "5.0",
    "conf.default.DGainL", "5.0",
    "conf.default.PGainR", "5.0",
    "conf.default.DGainR", "5.0",
    "conf.default.leftWheelID", "1",
    "conf.default.rightWheelID", "0",
    "conf.default.radiusOfLeftWheel", "0.018",
    "conf.default.radiusOfRightWheel", "0.018",
    "conf.default.lengthOfAxle", "0.078",
    "conf.default.radiusOfBodyArea", "0.15",
    "conf.default.torqueSensitivity", "1.0",
    "conf.default.torqueConstant", "1.0",
    // Widget
    "conf.__widget__.PGainL", "text",
    "conf.__widget__.DGainL", "text",
    "conf.__widget__.PGainR", "text",
    "conf.__widget__.DGainR", "text",
    "conf.__widget__.leftWheelID", "text",
    "conf.__widget__.rightWheelID", "text",
    "conf.__widget__.radiusOfLeftWheel", "text",
    "conf.__widget__.radiusOfRightWheel", "text",
    "conf.__widget__.lengthOfAxle", "text",
    "conf.__widget__.radiusOfBodyArea", "text",
    "conf.__widget__.torqueSensitivity", "text",
    "conf.__widget__.torqueConstant", "text",
    // Constraints
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
DataConversionRTC::DataConversionRTC(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_model_angleIn("model_angle", m_model_angle),
    m_target_velocityIn("target_velocity", m_target_velocity),
    m_model_torqueOut("model_torque", m_model_torque),
    m_robot_commandOut("robot_command", m_robot_command)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
DataConversionRTC::~DataConversionRTC()
{
}



RTC::ReturnCode_t DataConversionRTC::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("model_angle", m_model_angleIn);
  addInPort("target_velocity", m_target_velocityIn);
  
  // Set OutPort buffer
  addOutPort("model_torque", m_model_torqueOut);
  addOutPort("robot_command", m_robot_commandOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("PGainL", m_PGainL, "5.0");
  bindParameter("DGainL", m_DGainL, "5.0");
  bindParameter("PGainR", m_PGainR, "5.0");
  bindParameter("DGainR", m_DGainR, "5.0");
  bindParameter("leftWheelID", m_leftWheelID, "1");
  bindParameter("rightWheelID", m_rightWheelID, "0");
  bindParameter("radiusOfLeftWheel", m_radiusOfLeftWheel, "0.018");
  bindParameter("radiusOfRightWheel", m_radiusOfRightWheel, "0.018");
  bindParameter("lengthOfAxle", m_lengthOfAxle, "0.078");
  bindParameter("radiusOfBodyArea", m_radiusOfBodyArea, "0.15");
  bindParameter("torqueSensitivity", m_torqueSensitivity, "1.0");
  bindParameter("torqueConstant", m_torqueConstant, "1.0");
  
  // </rtc-template>

  //Jn
  m_model_torque.data.length(Dof);
  m_robot_command.data.length(4);
  //I

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t DataConversionRTC::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataConversionRTC::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataConversionRTC::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t DataConversionRTC::onActivated(RTC::UniqueId ec_id)
{
//---[add]---------------------------------
	RTC_INFO(("on Activated"));
	//std::cout << "on Activated" << std::endl;

	// Update INPORT 
	m_model_angleIn.update();
	m_target_velocityIn.update();

//--get Model Parameter from Configulation
	Body.leftID = m_leftWheelID; // jointID of left wheel
	Body.rightID = m_rightWheelID; // jointID of right wheel
	Body.wheelRadiusLeft = m_radiusOfLeftWheel; // Wheel Radius(Left)
	Body.wheelRadiusRight = m_radiusOfRightWheel; // Wheel Radius(Right)
	Body.bodyRadius = 0.5 * m_lengthOfAxle; // Body Radius (half of length of Wheel Axle)
	Body.length = m_radiusOfBodyArea; // length from center of Body to corner of Body

	RTC_DEBUG(("Body.leftID: %d   Body.rightID: %d", Body.leftID, Body.rightID));	
	RTC_DEBUG(("Body.wheelRadius[L]: %f Body.wheelRadius[R]: %f   Body.bodyRadius: %f",Body.wheelRadiusLeft, Body.wheelRadiusRight, Body.bodyRadius));	

  // set previous wheel angular velocity's data 
	for(int i=0; i < Dof; i++)  	qOld[i] = 0.0;  // initialize to current wheel's angle value

  // initialize SWITCH
	InitializeFlag = false;

	//-- initialize Time data
	sec = nsec = 0;	
	TimeStampOld = TimeStamp = 0.0;

//---[add]---------------------------------

	const char* output = "1kok";	//1kok:~
	memcpy(&(m_robot_command.data[0]),output,4);
	m_robot_commandOut.write();

  return RTC::RTC_OK;
}


RTC::ReturnCode_t DataConversionRTC::onDeactivated(RTC::UniqueId ec_id)
{
	const char* output = "1kok";	//1kok:~
	memcpy(&(m_robot_command.data[0]),output,4);
	m_robot_commandOut.write();

  return RTC::RTC_OK;
}

RTC::ReturnCode_t DataConversionRTC::onExecute(RTC::UniqueId ec_id)
{
//---[add]---------------------------------
//  std::cout << " ==================================== [onExecute] ====================================" << std::endl;

/*********************************************************************************************************/
//  main action
/*********************************************************************************************************/

	if(m_model_angleIn.isNew()){
		//-- get Previous Wheel Angle Data 
		m_model_angleIn.read();

		//-- set Previous Wheel Angle Data
		double q[Dof];
		q[Body.leftID] = m_model_angle.data[Body.leftID];
		q[Body.rightID] = m_model_angle.data[Body.rightID];

		//-- get time stamp data 
		sec = m_model_angle.tm.sec;
		nsec = m_model_angle.tm.nsec;

		TimeStamp = sec + 1.0e-9*nsec;

		//-- initialize 
		if (!InitializeFlag) {
			InitializeFlag = true;
		} else {
			//-- calc timestep value
			timeStep = TimeStamp - TimeStampOld;  

			//-- calc Wheel Angular Velocity's Datas by previous angle data
			double dq[Dof];
			dq[Body.leftID] = (q[Body.leftID] - qOld[Body.leftID]) / timeStep;
			dq[Body.rightID] = (q[Body.rightID] - qOld[Body.rightID]) / timeStep;

			//-- calc TORQUE  & set data to OUTPORT
			double q_ref[Dof], dq_ref[Dof];

			//-- action if Target Velocity datas are given 
			if(m_target_velocityIn.isNew()){

				//*** read Target Velocity data
				m_target_velocityIn.read();

				double TransVel,RotVel,Rho;
				TransVel = m_target_velocity.data.vx;   // [m/s]
				RotVel = m_target_velocity.data.va;    //  [rad/s]

				if (RotVel == 0.0) { 
					//-- calc [Target] Angular Velocity Dat
					dq_ref[Body.leftID] = TransVel / Body.wheelRadiusLeft * m_torqueConstant;
					dq_ref[Body.rightID] = TransVel / Body.wheelRadiusRight * m_torqueConstant;
				} else {
					Rho = TransVel / RotVel;        // curvature radius
					//-- calc [Target] Angular Velocity Dat
					dq_ref[Body.leftID] = (Rho - Body.bodyRadius) * RotVel / Body.wheelRadiusLeft * m_torqueConstant;
					dq_ref[Body.rightID] = (Rho + Body.bodyRadius) * RotVel / Body.wheelRadiusRight * m_torqueConstant;
				}

				//-- calc [Target] Angle Data
				q_ref[Body.leftID] = q[Body.leftID] + 0.5 * (dq_ref[Body.leftID] + dq[Body.leftID]) * timeStep;
				q_ref[Body.rightID] = q[Body.rightID] + 0.5 * (dq_ref[Body.rightID] + dq[Body.rightID]) * timeStep;

				//-- calc TORQUE value    
				m_model_torque.data[Body.leftID] = ( (q_ref[Body.leftID] - q[Body.leftID]) * m_PGainL + (dq_ref[Body.leftID] - dq[Body.leftID]) * m_DGainL ) * m_torqueSensitivity;
				m_model_torque.data[Body.rightID] = ( (q_ref[Body.rightID] - q[Body.rightID]) * m_PGainR + (dq_ref[Body.rightID] - dq[Body.rightID]) * m_DGainR ) * m_torqueSensitivity;

				RTC_DEBUG((" Pgain(L): %f Pgain(R): %f Dgain(L): %f Dgain(R): %f", m_PGainL, m_PGainR, m_DGainL, m_DGainR));;
				RTC_DEBUG((" TransVel: %f RotVel: %f", TransVel, RotVel));
				RTC_DEBUG((" q_ref(L): %f q(L): %f [q_ref-q](L): %f dq_ref(L): %f dq(L): %f [dq_ref-dq](L): %f", q_ref[Body.leftID], q[Body.leftID], q_ref[Body.leftID] - q[Body.leftID], dq_ref[Body.leftID], dq[Body.leftID], dq_ref[Body.leftID] - dq[Body.leftID]));	
				RTC_DEBUG((" q_ref(R): %f q(R): %f [q_ref-q](R): %f dq_ref(R): %f dq(R): %f [dq_ref-dq](R): %f", q_ref[Body.rightID], q[Body.rightID], q_ref[Body.rightID] - q[Body.rightID], dq_ref[Body.rightID], dq[Body.rightID], dq_ref[Body.rightID] - dq[Body.rightID]));
				RTC_DEBUG((" Ptorque(L): %f Dtorque(L): %f", (q_ref[Body.leftID] - q[Body.leftID]) * m_PGainL, (dq_ref[Body.leftID] - dq[Body.leftID]) * m_DGainL));
				RTC_DEBUG((" Ptorque(R): %f Dtorque(R): %f", (q_ref[Body.rightID] - q[Body.rightID]) * m_PGainR, (dq_ref[Body.rightID] - dq[Body.rightID]) * m_DGainR));

			} else {  
				m_model_torque.data[Body.leftID] = 0.0;
				m_model_torque.data[Body.rightID] = 0.0;
			}

	RTC_DEBUG(("[DataConversionRTC Torque] Time = %f[s]  Torque(L,R) = (%f, %f)", TimeStamp, m_model_torque.data[Body.leftID], m_model_torque.data[Body.rightID]));

			// set TIME
			m_model_torque.tm.sec = sec;
			m_model_torque.tm.nsec = nsec;
			//*** set Torque data to OutPort
			m_model_torqueOut.write();

		} // enod of if (initialize)

	//*** set Current data -> Previous data
		qOld[Body.leftID] = q[Body.leftID];
		qOld[Body.rightID] = q[Body.rightID];
		TimeStampOld = TimeStamp;

	}

//---[add]---------------------------------

/*********************************************************************************************************/
//  @̓ݒ@Jn
/*********************************************************************************************************/

	//*** read Target Velocity data
	m_target_velocityIn.read();

	double TransVel,RotVel;
	TransVel = m_target_velocity.data.vx;   // [m/s]
	RotVel = m_target_velocity.data.va;    //  [rad/s]

	const char* output;
	double RotVel_deg = RotVel * 180 / 3.141592;	//1bŉ]px[deg]Cv肪
	
	//ix[
	if(TransVel == 0.00){
		if(-90 <= RotVel_deg && RotVel_deg < -60){output = "2dok";}
		else if(-60 <= RotVel_deg && RotVel_deg < -30){output = "3dok";}
		else if(-30 <= RotVel_deg && RotVel_deg < 0){output = "4dok";}
		else if(RotVel_deg == 0.00){output = "1kok";}
		else if(0 < RotVel_deg && RotVel_deg <= 30){output = "4fok";}
		else if(30 < RotVel_deg && RotVel_deg <= 60){output = "3fok";}
		else if(60 < RotVel_deg && RotVel_deg <= 90){output = "2fok";}
	}

	//ixvX
	else if(0.00 < TransVel && TransVel < 0.20){
		if(-90 <= RotVel_deg && RotVel_deg < -60){output = "2dok";}
		else if(-60 <= RotVel_deg && RotVel_deg < -20){output = "5cok";}
		else if(-20 <= RotVel_deg && RotVel_deg <= 20){output = "5aok";}
		else if(20 < RotVel_deg && RotVel_deg <= 60){output = "5eok";}
		else if(60 < RotVel_deg && RotVel_deg <= 90){output = "2fok";}
	}
	else if(0.20 <= TransVel){
		if(-90 <= RotVel_deg && RotVel_deg < -60){output = "2dok";}
		else if(-60 <= RotVel_deg && RotVel_deg < -20){output = "4cok";}
		else if(-20 <= RotVel_deg && RotVel_deg <= 20){output = "4aok";}
		else if(20 < RotVel_deg && RotVel_deg <= 60){output = "4eok";}
		else if(60 < RotVel_deg && RotVel_deg <= 90){output = "2fok";}
	}
	//ix}CiX
	else if(0.00 > TransVel && TransVel > -0.20){
		if(-90 <= RotVel_deg && RotVel_deg < -60){output = "2hok";}
		else if(-60 <= RotVel_deg && RotVel_deg < -20){output = "5gok";}
		else if(-20 <= RotVel_deg && RotVel_deg <= 20){output = "5bok";}
		else if(20 < RotVel_deg && RotVel_deg <= 60){output = "5iok";}
		else if(60 < RotVel_deg && RotVel_deg <= 90){output = "2jok";}
	}
	else if(-0.20 >= TransVel){
		if(-90 <= RotVel_deg && RotVel_deg < -60){output = "2hok";}
		else if(-60 <= RotVel_deg && RotVel_deg < -20){output = "4gok";}
		else if(-20 <= RotVel_deg && RotVel_deg <= 20){output = "4bok";}
		else if(20 < RotVel_deg && RotVel_deg <= 60){output = "4iok";}
		else if(60 < RotVel_deg && RotVel_deg <= 90){output = "2jok";}
	}
	memcpy(&(m_robot_command.data[0]),output,4);
	m_robot_commandOut.write();
	std::cout << "m_robot_command.data[0][1][2][3] = " << m_robot_command.data[0] << m_robot_command.data[1] << m_robot_command.data[2] << m_robot_command.data[3] << std::endl;

/*********************************************************************************************************/
//  @̓ݒ@I
/*********************************************************************************************************/

  return RTC::RTC_OK;
}
/*
RTC::ReturnCode_t DataConversionRTC::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataConversionRTC::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataConversionRTC::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataConversionRTC::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataConversionRTC::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void DataConversionRTCInit(RTC::Manager* manager)
  {
    coil::Properties profile(dataconversionrtc_spec);
    manager->registerFactory(profile,
                             RTC::Create<DataConversionRTC>,
                             RTC::Delete<DataConversionRTC>);
  }
  
};


