// -*- C++ -*-
/*!
 * @file  DataIntegrationRTC.cpp
 * @brief DataIntegrationRTC
 * @date $Date$
 *
 * $Id$
 */

#include "DataIntegrationRTC.h"

// Module specification
// <rtc-template block="module_spec">
static const char* dataintegrationrtc_spec[] =
  {
    "implementation_id", "DataIntegrationRTC",
    "type_name",         "DataIntegrationRTC",
    "description",       "DataIntegrationRTC",
    "version",           "1.0.0",
    "vendor",            "TakahashiNaoki",
    "category",          "Three-WheeledMobileRobot",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
DataIntegrationRTC::DataIntegrationRTC(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_in_angleIn("in_angle", m_in_angle),
    m_in_color_imageIn("in_color_image", m_in_color_image),
    m_in_range_sensorIn("in_range_sensor", m_in_range_sensor),
    m_in_touch_sensorIn("in_touch_sensor", m_in_touch_sensor),
    m_out_angleOut("out_angle", m_out_angle),
    m_out_color_imageOut("out_color_image", m_out_color_image),
    m_out_range_sensorOut("out_range_sensor", m_out_range_sensor),
    m_out_touch_sensorOut("out_touch_sensor", m_out_touch_sensor)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
DataIntegrationRTC::~DataIntegrationRTC()
{
}



RTC::ReturnCode_t DataIntegrationRTC::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("in_angle", m_in_angleIn);
  addInPort("in_color_image", m_in_color_imageIn);
  addInPort("in_range_sensor", m_in_range_sensorIn);
  addInPort("in_touch_sensor", m_in_touch_sensorIn);
  
  // Set OutPort buffer
  addOutPort("out_angle", m_out_angleOut);
  addOutPort("out_color_image", m_out_color_imageOut);
  addOutPort("out_range_sensor", m_out_range_sensorOut);
  addOutPort("out_touch_sensor", m_out_touch_sensorOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  //Jn
  m_in_angle.data.length(2);
  m_out_angle.data.length(2);
  m_in_angle.data[0] = m_in_angle.data[1] = 0;
  m_out_angle.data[0] = m_out_angle.data[1] = 0;
  m_out_touch_sensor.data = 0;
  //I

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t DataIntegrationRTC::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataIntegrationRTC::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataIntegrationRTC::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataIntegrationRTC::onActivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataIntegrationRTC::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t DataIntegrationRTC::onExecute(RTC::UniqueId ec_id)
{
	if(m_in_angleIn.isNew()){

		m_in_angleIn.read();

		m_out_angle.data[0] = m_in_angle.data[0];	//0:right
		m_out_angle.data[1] = m_in_angle.data[1];	//1:left

		m_out_angle.tm.sec = m_in_angle.tm.sec;
		m_out_angle.tm.nsec = m_in_angle.tm.nsec;

		m_out_angleOut.write();
	}

	if(m_in_color_imageIn.isNew()){
		int i;
		m_in_color_imageIn.read();
		int image_datasize = (int)m_in_color_image.data.length();
		m_out_color_image.data.length(image_datasize);
		
		for(i=0;i<image_datasize;i++){
			m_out_color_image.data[i] = m_in_color_image.data[i];
		}
		m_out_color_imageOut.write();
	}

	if(m_in_range_sensorIn.isNew()){
		int j;
		m_in_range_sensorIn.read();	//ZT̓͂̓ǂݍ
		int range_datasize = (int)m_in_range_sensor.data.length();	//ZT̃f[^mF
		m_out_range_sensor.data.length(range_datasize);

		for(j=0;j<range_datasize;j++){
			m_out_range_sensor.data[j] = (long)(1000 * m_in_range_sensor.data[j]);	//mm
		}
		m_out_range_sensorOut.write();
	}

	m_out_touch_sensor.data = 0;
	if(m_in_touch_sensorIn.isNew()){
		int k;
		m_in_touch_sensorIn.read();
		int touch_datasize = (int)m_in_touch_sensor.data.length();
		for(k=0;k<touch_datasize;k++){
			if(m_in_touch_sensor.data[k] == '1'){
				m_out_touch_sensor.data = 1;
			}
		}
		m_out_touch_sensorOut.write();
	}

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t DataIntegrationRTC::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataIntegrationRTC::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataIntegrationRTC::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataIntegrationRTC::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataIntegrationRTC::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void DataIntegrationRTCInit(RTC::Manager* manager)
  {
    coil::Properties profile(dataintegrationrtc_spec);
    manager->registerFactory(profile,
                             RTC::Create<DataIntegrationRTC>,
                             RTC::Delete<DataIntegrationRTC>);
  }
  
};


